#include "float.h"
#include "aspen_devel.h"


#define DEFAULT_BLOCK_X		8


#if GPU_ARCH == Tesla
	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	1
	#define	USE_RESTRICT	0
	#define	USE_LDG		0

#else
#if GPU_ARCH == Fermi
	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	1
	#define	USE_RESTRICT	0
	#define	USE_LDG		0

#else
#if GPU_ARCH == Kepler
	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	0
	#define	USE_RESTRICT	1
	#define	USE_LDG		0

#else
#if GPU_ARCH == Kepler2
	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	0
	#define	USE_RESTRICT	0
	#define	USE_LDG		1

#else
#if GPU_ARCH == Maxwell
	#define	USE_VOLATILE	0
	#define	USE_TEXTURE	0
	#define	USE_RESTRICT	0
	#define USE_LDG		1

#else
#if GPU_ARCH == Maxwell2
	#define	USE_VOLATILE	0
	#define	USE_TEXTURE	0
	#define	USE_RESTRICT	0
	#define USE_LDG		1

#else
	error !!!
#endif // Maxwell2
#endif // Maxwell
#endif // Kepler2
#endif // Kepler
#endif // Fermi
#endif // Tesla

#include "aspen_devel_postdef.h"

#if USE_TEXTURE
	DEF_TEXTURE ( texture_X, X_ );
	#define	__TRICK_offsetX	,offsetX
#else
	#define	__TRICK_offsetX
#endif


#define	SUMUP_PREFIX		_NAME_1PARAMS_(SGEMVN,GPU_PREFIX)
#define	get_opt_parameter	_NAME_1PARAMS_(SUMUP_PREFIX,Get_Opt_Parameter)
#define	SGEMV_n_STUB		_NAME_1PARAMS_(SGEMV_n,GPU_PREFIX)


#include "../tuning/CURRENT_GPU"

#define	ALL_KERNELS	0

#if CURRENT_GPU==GPU_ARCH
	#include "../template/sumup_2D_template.incl"
	#include "../template/gemv_n_template.incl"
	#include "../tuning/sgemvn_opt4.h"
#endif


#if defined(__CY__)
   #define	_CASE__(x, ... ) \
			_CASE_( x, _BLOCKSIZE_##x, _UX_##x, _MULTI_##x )
#else
//   #define	_BLOCKSIZE_(x)		( (((x)/2000)+1)*32 )
   #define	_BLOCKSIZE_(x)		( ((x)<100000) ? ((((x)/1000)+2)*16) : 16 )
   #define	_UX_(x)			( (x%100) )
   #define	_CASE__(x, ... ) \
			_CASE_( x, _BLOCKSIZE_(x), _UX_(x), 1 )
#endif


#define _CASE_(N,X,Y,MULTI)	\
		case N: \
			API(GEMVN_2D_host) \
				< GPU_ARCH, scalar_t, X, Y > \
				( alpha, beta, a, x, y, lda, \
					 n1, n2, MULTI __TRICK_offsetX ); break


extern "C" void
API(SGEMV_n_STUB) ( int blk,
		const scalar_t alpha, const scalar_t beta,
		const scalar_t * a, const scalar_t * x,
		scalar_t * y,
                const int lda,
		const int n1, const int n2 )
{

#if USE_TEXTURE
	size_t offsetX = 0;
	SETUP_TEXTURE ( texture_X, x, n2, offsetX );
#endif

#if CURRENT_GPU==GPU_ARCH
	int BLK = blk;
	int n = n2;

	#include "../tuning/sgemvn-auto2.h"

	int m = (blk<=0) ? BLK : blk;

	switch ( m ) {

#if defined(KERNEL_1) || ALL_KERNELS
        _CASE__(     1); // *
#endif
#if defined(KERNEL_101) || ALL_KERNELS
        _CASE__(   101); // *
#endif
#if defined(KERNEL_201) || ALL_KERNELS
        _CASE__(   201); // *
#endif
#if defined(KERNEL_2) || ALL_KERNELS
        _CASE__(     2); // *
#endif
#if defined(KERNEL_102) || ALL_KERNELS
        _CASE__(   102); // *
#endif
#if defined(KERNEL_202) || ALL_KERNELS
        _CASE__(   202); // *
#endif
#if defined(KERNEL_3) || ALL_KERNELS
        _CASE__(     3); // *
#endif
#if defined(KERNEL_103) || ALL_KERNELS
        _CASE__(   103); // *
#endif
#if defined(KERNEL_203) || ALL_KERNELS
        _CASE__(   203); // *
#endif
#if defined(KERNEL_4) || ALL_KERNELS
        _CASE__(     4); // *
#endif
#if defined(KERNEL_104) || ALL_KERNELS
        _CASE__(   104); // *
#endif
#if defined(KERNEL_204) || ALL_KERNELS
        _CASE__(   204); // *
#endif
#if defined(KERNEL_5) || ALL_KERNELS
        _CASE__(     5); // *
#endif
#if defined(KERNEL_105) || ALL_KERNELS
        _CASE__(   105); // *
#endif
#if defined(KERNEL_205) || ALL_KERNELS
        _CASE__(   205); // *
#endif
#if defined(KERNEL_6) || ALL_KERNELS
        _CASE__(     6); // *
#endif
#if defined(KERNEL_106) || ALL_KERNELS
        _CASE__(   106); // *
#endif
#if defined(KERNEL_206) || ALL_KERNELS
        _CASE__(   206); // *
#endif
#if defined(KERNEL_7) || ALL_KERNELS
        _CASE__(     7); // *
#endif
#if defined(KERNEL_107) || ALL_KERNELS
        _CASE__(   107); // *
#endif
#if defined(KERNEL_207) || ALL_KERNELS
        _CASE__(   207); // *
#endif
#if defined(KERNEL_8) || ALL_KERNELS
        _CASE__(     8); // *
#endif
#if defined(KERNEL_108) || ALL_KERNELS
        _CASE__(   108); // *
#endif
#if defined(KERNEL_208) || ALL_KERNELS
        _CASE__(   208); // *
#endif
#if defined(KERNEL_9) || ALL_KERNELS
        _CASE__(     9); // *
#endif
#if defined(KERNEL_109) || ALL_KERNELS
        _CASE__(   109); // *
#endif
#if defined(KERNEL_209) || ALL_KERNELS
        _CASE__(   209); // *
#endif
#if defined(KERNEL_10) || ALL_KERNELS
        _CASE__(    10); // *
#endif
#if defined(KERNEL_110) || ALL_KERNELS
        _CASE__(   110); // *
#endif
#if defined(KERNEL_210) || ALL_KERNELS
        _CASE__(   210); // *
#endif
#if defined(KERNEL_11) || ALL_KERNELS
        _CASE__(    11); // *
#endif
#if defined(KERNEL_111) || ALL_KERNELS
        _CASE__(   111); // *
#endif
#if defined(KERNEL_211) || ALL_KERNELS
        _CASE__(   211); // *
#endif
#if defined(KERNEL_12) || ALL_KERNELS
        _CASE__(    12); // *
#endif
#if defined(KERNEL_112) || ALL_KERNELS
        _CASE__(   112); // *
#endif
#if defined(KERNEL_212) || ALL_KERNELS
        _CASE__(   212); // *
#endif
#if defined(KERNEL_13) || ALL_KERNELS
        _CASE__(    13); // *
#endif
#if defined(KERNEL_113) || ALL_KERNELS
        _CASE__(   113); // *
#endif
#if defined(KERNEL_14) || ALL_KERNELS
        _CASE__(    14); // *
#endif
#if defined(KERNEL_114) || ALL_KERNELS
        _CASE__(   114); // *
#endif
#if defined(KERNEL_15) || ALL_KERNELS
        _CASE__(    15); // *
#endif
#if defined(KERNEL_115) || ALL_KERNELS
        _CASE__(   115); // *
#endif
#if defined(KERNEL_16) || ALL_KERNELS
        _CASE__(    16); // *
#endif
#if defined(KERNEL_116) || ALL_KERNELS
        _CASE__(   116); // *
#endif

#if defined(KERNEL_1001) || ALL_KERNELS
        _CASE__(  1001); // *
#endif
#if defined(KERNEL_1101) || ALL_KERNELS
        _CASE__(  1101); // *
#endif
#if defined(KERNEL_1002) || ALL_KERNELS
        _CASE__(  1002); // *
#endif
#if defined(KERNEL_1102) || ALL_KERNELS
        _CASE__(  1102); // *
#endif
#if defined(KERNEL_1003) || ALL_KERNELS
        _CASE__(  1003); // *
#endif
#if defined(KERNEL_1103) || ALL_KERNELS
        _CASE__(  1103); // *
#endif

#if defined(KERNEL_2001) || ALL_KERNELS
        _CASE__(  2001); // *
#endif
#if defined(KERNEL_2101) || ALL_KERNELS
        _CASE__(  2101); // *
#endif
#if defined(KERNEL_2002) || ALL_KERNELS
        _CASE__(  2002); // *
#endif
#if defined(KERNEL_2102) || ALL_KERNELS
        _CASE__(  2102); // *
#endif
#if defined(KERNEL_2003) || ALL_KERNELS
        _CASE__(  2003); // *
#endif
#if defined(KERNEL_2103) || ALL_KERNELS
        _CASE__(  2103); // *
#endif
#if defined(KERNEL_2004) || ALL_KERNELS
        _CASE__(  2004); // *
#endif
#if defined(KERNEL_2104) || ALL_KERNELS
        _CASE__(  2104); // *
#endif
#if defined(KERNEL_2005) || ALL_KERNELS
        _CASE__(  2005); // *
#endif
#if defined(KERNEL_2105) || ALL_KERNELS
        _CASE__(  2105); // *
#endif
#if defined(KERNEL_2006) || ALL_KERNELS
        _CASE__(  2006); // *
#endif
#if defined(KERNEL_2106) || ALL_KERNELS
        _CASE__(  2106); // *
#endif
#if defined(KERNEL_2007) || ALL_KERNELS
        _CASE__(  2007); // *
#endif
#if defined(KERNEL_2008) || ALL_KERNELS
        _CASE__(  2008); // *
#endif

#if defined(KERNEL_3001) || ALL_KERNELS
        _CASE__(  3001); // *
#endif
#if defined(KERNEL_3101) || ALL_KERNELS
        _CASE__(  3101); // *
#endif

#if defined(KERNEL_4001) || ALL_KERNELS
        _CASE__(  4001); // *
#endif
#if defined(KERNEL_4002) || ALL_KERNELS
        _CASE__(  4002); // x
#endif
#if defined(KERNEL_4003) || ALL_KERNELS
        _CASE__(  4003); // x
#endif
#if defined(KERNEL_4004) || ALL_KERNELS
        _CASE__(  4004); // x
#endif
#if defined(KERNEL_4005) || ALL_KERNELS
        _CASE__(  4005); // x
#endif

#if defined(KERNEL_5001) || ALL_KERNELS
        _CASE__(  5001); // *
#endif

#if defined(KERNEL_6001) || ALL_KERNELS
        _CASE__(  6001); // *
#endif
#if defined(KERNEL_6002) || ALL_KERNELS
        _CASE__(  6002); // xx
#endif
#if defined(KERNEL_6003) || ALL_KERNELS
        _CASE__(  6003); // xx
#endif
#if defined(KERNEL_6004) || ALL_KERNELS
        _CASE__(  6004); // xx
#endif

#if defined(KERNEL_7001) || ALL_KERNELS
        _CASE__(  7001); // *
#endif

#if defined(KERNEL_8001) || ALL_KERNELS
        _CASE__(  8001); // xx
#endif
#if defined(KERNEL_8002) || ALL_KERNELS
        _CASE__(  8002); // xx
#endif
#if defined(KERNEL_8003) || ALL_KERNELS
        _CASE__(  8003); // xx
#endif

#if defined(KERNEL_9001) || ALL_KERNELS
        _CASE__(  9001); // *
#endif

#if defined(KERNEL_10001) || ALL_KERNELS
        _CASE__(  10001); // xx
#endif
#if defined(KERNEL_10002) || ALL_KERNELS
        _CASE__(  10002); // xx
#endif

#if defined(KERNEL_11001) || ALL_KERNELS
        _CASE__(  11001); // *
#endif

#if defined(KERNEL_12001) || ALL_KERNELS
        _CASE__(  12001); // xx
#endif
#if defined(KERNEL_12002) || ALL_KERNELS
        _CASE__(  12002); // xx
#endif

#if defined(KERNEL_13001) || ALL_KERNELS
        _CASE__(  13001); // *
#endif

#if defined(KERNEL_14001) || ALL_KERNELS
        _CASE__(  14001); // xx
#endif
#if defined(KERNEL_14002) || ALL_KERNELS
        _CASE__(  14002); // xx
#endif

#if defined(KERNEL_16001) || ALL_KERNELS
        _CASE__(  16001); // xx
#endif
#if defined(KERNEL_18001) || ALL_KERNELS
        _CASE__(  18001); // xx
#endif
#if defined(KERNEL_20001) || ALL_KERNELS
        _CASE__(  20001); // xx
#endif
#if defined(KERNEL_22001) || ALL_KERNELS
        _CASE__(  22001); // xx
#endif

#if defined(KERNEL_100006) || ALL_KERNELS
        _CASE__( 100006); // xx
#endif
#if defined(KERNEL_100007) || ALL_KERNELS
        _CASE__( 100007); // xx
#endif
#if defined(KERNEL_100008) || ALL_KERNELS
        _CASE__( 100008); // xx
#endif
#if defined(KERNEL_100009) || ALL_KERNELS
        _CASE__( 100009); // xx
#endif
#if defined(KERNEL_100010) || ALL_KERNELS
        _CASE__( 100010); // xx
#endif
#if defined(KERNEL_100011) || ALL_KERNELS
        _CASE__( 100011); // xx
#endif
#if defined(KERNEL_100012) || ALL_KERNELS
        _CASE__( 100012); // xx
#endif
#if defined(KERNEL_100013) || ALL_KERNELS
        _CASE__( 100013); // xx
#endif
#if defined(KERNEL_100014) || ALL_KERNELS
        _CASE__( 100014); // xx
#endif
#if defined(KERNEL_100015) || ALL_KERNELS
        _CASE__( 100015); // xx
#endif
#if defined(KERNEL_100016) || ALL_KERNELS
        _CASE__( 100016); // xx
#endif
#if defined(KERNEL_100106) || ALL_KERNELS
        _CASE__( 100106); // xx
#endif
#if defined(KERNEL_100107) || ALL_KERNELS
        _CASE__( 100107); // xx
#endif
#if defined(KERNEL_100108) || ALL_KERNELS
        _CASE__( 100108); // xx
#endif
#if defined(KERNEL_100109) || ALL_KERNELS
        _CASE__( 100109); // xx
#endif
	default:
		fprintf ( stderr,"Not proper NB %d ", m );
		break;
	}
#else
	fprintf( stderr,
		"Caution, ASPEN is tuned up for %s.\n",
		ASPEN_GPU_Names(CURRENT_GPU) );
#endif

#if USE_TEXTURE
	cudaUnbindTexture ( texture_X );
#endif

}

#undef	_CASE_

