#!/bin/bash

prolog()
{
	if [ x${HOST} = x ]; then
		HOST=`/bin/hostname`
	fi

	../src/get_dev_info $ASPEN_GPU_ID > DEV_INFO

	DEVICE=`awk '/DEVICE=/{ print $2 }' DEV_INFO`
	MP=`awk '/MP=/{ print $2 }' DEV_INFO`
	CG=`awk '/CG=/{ print $2 }' DEV_INFO`
	MAXDIM=`awk '/MAXDIM2/{ print $2 }' DEV_INFO`

	\rm DEV_INFO

	if [ -e CURRENT_GPU ]; then
	CG_=`awk '{print $3*1}' CURRENT_GPU`
	if [ $CG -ne $CG_ ]; then
		\rm CURRENT_GPU
		echo '#define CURRENT_GPU '$CG > CURRENT_GPU
	fi
	else
		\rm CURRENT_GPU
		echo '#define CURRENT_GPU '$CG > CURRENT_GPU
	fi

	if [ $CG -lt 300 ]; then
		MAX_MULTI=8
		MM_PATTERN="1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16"
		REG=`expr 32 "*" 1024`
	else
		MAX_MULTI=16
		MM_PATTERN="1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32"
		REG=`expr 64 "*" 1024`
	fi


	BENCHCODE=../bench/test2-sgemv-t
	CORES=`cat /proc/cpuinfo | awk 'BEGIN{c=1}/cores/{c=$4}END{print c}'`
	if [ $CORES -gt 1 ]; then
		EXEC='' #'/usr/bin/numactl --physcpubind=0 '
	else
		EXEC=''
	fi

	IN_DATA=IN-a
	FIT_INIT='a=20; o1=100; o2=1; o3=0.3;'

	if [ x${DEVICE} = xunknown ]; then
		echo '**** CAUTION !!! ****'
		echo 'CUDA environment does not match to your GPU'
		echo 'Check the installed NVIDIA driver and the version of CUDA!'
		exit 1
	fi

	/bin/sh ./touch_header.sh
}

echo_Message()
{
	echo '<--/****************************************'
	echo 'Automatic Performance Tuning for SGEMV -T'
	date
	echo -n 'Host on '; echo `hostname`
	echo 'Device is '$DEVICE
	if [ x${1} = xbegin ]; then
		echo 'Start.'
	fi
	if [ x${1} = xend ]; then
		echo 'Successfully completed.'
	fi
	if [ x${1} = xterminate ]; then
		echo 'Terminated unfortunately.'
	fi
	echo '****************************************/-->'
}

update_MYBLAS()
{
	/bin/sh ./get_reguse.sh $CG GEMVT_kernel \
		../src/sgemv_t-X_template \
		../src/sgemv_t-X_template \
		../src/sgemv_t-X_template \
		../src/sgemv_t-X_template \
		../src/sgemv_t-X_template \
		../src/sgemv_t-X_template \
		| \
	sort -k 4 > log-reg-all

	export ASPEN_VERIFY=1
	cd ../src
	\rm sgemv_t.cu_o
	\rm sgemv_t-*.cu_o
	make -j4
	cd ../bench
	\rm test-f.o
	make test-s
	cd ../tuning
	unset ASPEN_VERIFY
}

initial_Phase()
{
	touch at-log
	\rm at-log

	if [ -d logs ]; then
		\rm -rf logs
	fi
}

init_IN_file()
{
	touch $IN_DATA
	\rm $IN_DATA
	echo 1 > $IN_DATA
	for N in \
		100 500 \
		1000 1100 1250 1500 1750 \
		2000 2200 2500 3000 3500 4000 \
		5000 6000 8000 10000
	do
		echo  $N >> $IN_DATA
	done
	N=12000
	while [ $N -lt $MAXDIM ]
	do
		echo $N >> $IN_DATA
		N=`expr $N "+" 2000`
	done
		echo $MAXDIM >> $IN_DATA 
	echo 0 >> $IN_DATA
}

init_IN_file2()
{
	touch $IN_DATA
	\rm $IN_DATA
	echo 1 > $IN_DATA
#	awk 'BEGIN{for(i=64;i<1024;i+=64)print i}' >> $IN_DATA
#	awk 'BEGIN{for(i=1024;i<2048;i+=16)print i}' >> $IN_DATA
#	for N in \
#		500 800 1000 1200 \
#		1274 1276 1278 1280 1282 1284 1286 \
#		1594 1596 1598 1600 1602 1604 1606 \
#		1800 \
#		1914 1916 1918 1920 1922 1924 1926 \
#		2000 \
#		2042 2044 2046 2048 2050 2052 2054 \
#		2800 \
#		2872 2874 2876 2888 2880 2882 2884 \
#		3000 \
#		3063 3066 3069 3072 3075 3078 3081 \
#		3200 \
#		3800 \
#		3834 3836 3838 3840 3842 3844 3846 \
#		4000 \
#		4084 4088 4092 4096 4100 4104 4108 \
#		4794 4796 4798 4800 4802 4804 4806 \
#		5000 5200 \
#		5800 6000 6200 \
#		6800 7000 7200 \
#		7800 8000 8200 \
#		8800 9000 9200
#	do
#		echo  $N >> $IN_DATA
#	done
#	N=10000
#	while [ $N -lt $MAXDIM ]
#	do
#		echo $N >> $IN_DATA
#		N=`expr $N "+" 2000`
#	done
	echo $MAXDIM | awk '{MAXDIM=$0; for(i=100;i<int(MAXDIM/100);i+=4)print i; exit; }' >> $IN_DATA;
	echo $MAXDIM | awk '{MAXDIM=$0; for(i=int(MAXDIM/100);i<int(MAXDIM/10);i+=32)print i; exit; }' >> $IN_DATA;
	echo $MAXDIM | awk '{MAXDIM=$0; for(i=int(MAXDIM/10);i<int(MAXDIM/4);i+=280)print i; exit; }' >> $IN_DATA;
	echo $MAXDIM | awk '{MAXDIM=$0; for(i=int(MAXDIM/4);i<int(MAXDIM/3);i+=1000)print i; exit; }' >> $IN_DATA;
	echo $MAXDIM | awk '{MAXDIM=$0; for(i=int(MAXDIM/3);i<int(MAXDIM/2);i+=2000)print i; exit; }' >> $IN_DATA;
	echo $MAXDIM | awk '{MAXDIM=$0; for(i=int(MAXDIM/2);i<int(MAXDIM/1);i+=2800)print i; exit; }' >> $IN_DATA;
		echo $MAXDIM >> $IN_DATA 
	echo 0 >> $IN_DATA
}

do_AT_sampler_SUB()
{
	IDX=$KERNEL_IDX

	UX_PATTERN=" 2 3 4 5 6 7 8"
	UNROLL_PATTERN=" 1 "

	echo "" > sgemvt-at-opt.c

	cat << EOS_S0 >> sgemvt-at-opt.c


long
do_best_kernel_SGEMVT(int blk, const scalar_t alpha, const scalar_t beta,
const scalar_t *a, const scalar_t *x, scalar_t *y,
int lda, int n1, int n2, int offsetX, double *ff)
{
switch (blk) {
EOS_S0

	KERNEL_IDX=$IDX
	for UMAX in $UX_PATTERN; do
	for UNROLL in $UNROLL_PATTERN; do
		KERNEL_IDX=`expr $KERNEL_IDX "+" 1`
		echo "case "$KERNEL_IDX":" >> sgemvt-at-opt.c
		echo "GEMVT_KERNEL < GPU_ARCH, scalar_t, "$BLOCK_SIZE", "$VMAX", "$UMAX", "$UNROLL" >" >> sgemvt-at-opt.c
		echo "( alpha, beta, a, x, y, lda, n1, n2, offsetX );" >> sgemvt-at-opt.c
		echo "break;" >> sgemvt-at-opt.c
	done
	done

	cat << EOS_S1 >> sgemvt-at-opt.c
}
return (long) blk;
}
EOS_S1

	update_MYBLAS >& /dev/null

cat << EOF_IN1 > IN-1
1
6000
-1
EOF_IN1

	KERNEL_IDX=$IDX
	for UMAX in $UX_PATTERN; do
	for UNROLL in $UNROLL_PATTERN; do
        if [ `expr $UMAX "*" $VMAX` -gt 1 ]; then

		KERNEL_IDX=`expr $KERNEL_IDX "+" 1`
		LOG=logs/log-MYSGEMV-T-$BLOCK_SIZE"x"$UMAX"x"$VMAX"x"$UNROLL-X-GTX560

		cat log-reg-all | awk '{if(($2=='$BLOCK_SIZE')&&($3=='$UMAX')&&($4=='$VMAX'))print $1}' > CX


		CX=`cat CX`
		M=`expr $UMAX "+" $BLOCK_SIZE`
		M=`expr $M "+" 1`
		M=`expr $M "*" $VMAX`
		MX=`expr $M "*" 4`
		BX=`expr $BLOCK_SIZE "*" $VMAX`

if [ $CG < 200 ]; then
		MM_PATTERN=`/bin/sh ./get_mult.sh $BX $CX $MX $CG | \
			awk '{ gsub(/\([0-9]*\)/,""); print; }'`
else
		MM_PATTERN=`/bin/sh ./get_mult48.sh $BX $CX $MX $CG | \
			awk '{ gsub(/\([0-9]*\)/,""); print; }'`
fi

		MM_LAST=`echo $MM_PATTERN | awk '{ print 1*$(NF); }'`
        if [ $MM_LAST -ge 4 ]; then

		echo "#BEGIN "$BLOCK_SIZE" "$UMAX" "$VMAX" "$UNROLL

		M=`expr $VMAX "*" $UMAX`
		M=`expr $M "*" $MP`
		M=`expr $M "*" $MM_LAST`

		echo '% CX='$CX' MM='$MM_PATTERN
		echo '% CYCLE='$M

		unset ASPEN_VERIFY
		$EXEC $BENCHCODE $IN_DATA $KERNEL_IDX \
			| awk '/ERROR/{exit}{print;fflush();}' \
			| tee $LOG
		echo 'CYCLE= '$M >> $LOG
		echo "#END "$BLOCK_SIZE" "$UMAX" "$VMAX" "$UNROLL

	fi

	fi
	done
	done
}

do_AT_sampler()
{
	echo 'First Sampling, this is a very heavy step.'
	echo 'It might take a couple of hours, please wait.'

	if [ ! -d logs ]; then
		mkdir logs 
	fi

	init_IN_file

	if [ $CG -lt 200 ]; then
		BLK_PATTERN="32 64 96 128 192 256"
		VX_PATTERN="1 2 3 4"
	else
	if [ $CG -lt 300 ]; then
		BLK_PATTERN="16 32 48 64 96 128 160 192 224 256"
		VX_PATTERN="1 2 3 4"
	else
		BLK_PATTERN="16 32 64 96 128 160 192 224 256 288 320 352 384 "
		VX_PATTERN="1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24"
	fi
	fi

	KERNEL_IDX=0
	for BLOCK_SIZE in $BLK_PATTERN; do
	for VMAX in $VX_PATTERN; do
		if [ `expr $BLOCK_SIZE "*" $VMAX` -ge 96  ]; then
		if [ `expr $BLOCK_SIZE "*" $VMAX` -le 384 ]; then

			do_AT_sampler_SUB

		fi
		fi
	done
	done

}

do_AT_sampler2()
{
	#echo 'Second Sampling, it might take a couple of minutes.'
	echo 'Second Sampling, in order to detect the frequency of'
	echo 'sawtooth-figured profile of the kernel function accurately,'
	echo 'it might take a couple of hours, too. Please wait again.'

	if [ ! -d logs ]; then
		mkdir logs 
	fi

	create_AT_sampler2

	cat at.data | \
	awk ' \
	/DIAG/{a=1;next;} \
	{if(a)print;} \
	' | \
	awk ' \
	/^[0-9]/{ \
		split($2,aa,":"); \
		print "parameter="aa[1]": "aa[2]": "aa[3]": "aa[4]; \
		system("/bin/sh exe-sgemvt-sub2 "aa[1]" "aa[2]" "aa[3]" "aa[4]" '$IN_DATA' 64 '$MP' -"); \
	} \
	'
}

do_Ranking()
{
	touch at.data
	\rm at.data

	echo 'Screening by the point-ranking'

	cat at-log | \
	awk '\
	BEGIN{ \
		RANK_NN=50;\
		getline f < "'$IN_DATA'"; \
		i=0; while ( getline f < "'$IN_DATA'" ) { \
			if ( f < 1 ) break; \
			NN = f; \
			N[i++] = NN; \
			for(j=0;j<RANK_NN;j++)P[NN":"j] = 0; \
			for(j=0;j<RANK_NN;j++)name[NN":"j] = ""; \
		} \
	} \
	/^#BEGIN /{ \
		thread_num = $2; \
		umax       = $3; \
		vmax       = $4; \
		unroll     = $5; \
#		print; \
		next; \
	} \
	/^#END /{ \
#		print; \
		next; \
	} \
	/N=/{ \
		NN = $2; \
		f  = $5; \
		for(j=0;j<RANK_NN;j++){ \
			if ( f >= P[NN":"j] ) { \
				for(k=RANK_NN-1;k>j;k--){ \
					P[NN":"k] = P[NN":"k-1]; \
					name[NN":"k] = name[NN":"k-1]; \
				} \
				P[NN":"j] = f; \
				name[NN":"j] = thread_num":"umax":"vmax":"unroll; \
				break; \
			} \
		} \
#		print; \
		next; \
	} \
	END { \
		# Ranking point !!
		POINT[0] =2000; \
		POINT[1] = 500; \
		POINT[2] = 300; \
		POINT[3] = 200; \
		POINT[4] = 100; \
		POINT[5] =  50; \
		POINT[6] =  40; \
		POINT[7] =  20; \
		POINT[8] =  10; \
		POINT[9] =   5; \
		for(j=10;j<RANK_NN;j++) POINT[j] = 1; \
		print "DIAGNOSTICS***"; \
		print "POINT BLOCK_SIZE:UMAX:VMAX:UNROLL"; \
		for(j=0;j<i;j++){ \
			NN = N[j]; \
			for(k=0;k<RANK_NN;k++) { \
				PPP[name[NN":"k]]+=POINT[k]; \
			} \
		} \
		for(i=0;i<RANK_NN;i++) { xx[i]=0; yy[i]=""; } \
		for( s in PPP ) { if ( s == "" ) continue; \
			for(i=0;i<RANK_NN;i++){ \
				if ( PPP[s] >= xx[i] ) break; \
			} \
			if(i<RANK_NN){ \
				for(j=RANK_NN-1;j>i;j--){ \
					xx[j]=xx[j-1]; \
					yy[j]=yy[j-1]; \
				} \
				xx[i] = PPP[s]; \
				yy[i] = s; \
			} \
		} \
		for(i=0;i<20;i++) { print xx[i]" "yy[i]; } \
	}' | tee at.data

}

create_GP_plot()
{
	touch bb.plot
	\rm bb.plot

#	cat << EOS_0 > bb.plot
#$FIT_INIT
#f(x)=(a)*(x*x)/(abs(o1**2)+((o2)+x)*(x));
#fit f(x) 'log' u 2:5 via o1,o2,a
#EOS_0
	cat << EOS_0 > bb.plot
$FIT_INIT
load 'param.plt';
B=CYCLE;
#if ( $CG >= 200 ) B=MP*UMAX*VMAX;
if ( B <= 11*MP ) B=1;
if ( B <= 0     ) B=1;
f(x)=(a)*(x*x)/(o1**2+((o2)+(o3)*floor(x/B)*B+(1.0-(o3))*x)*(x));
#f(x)=(a)*(x*x)/(o1**2+((o2)+(o3)*floor(x/B)+x)*(x));
set xr[0:2048]
if ( B >  1 ) fit f(x) 'log' u 2:5 via o1,o2,o3,a
if ( B == 1 ) fit f(x) 'log' u 2:5 via o1,o2,a
!rm fit.log
set xr[2048:4096]
if ( B >  1 ) fit f(x) 'log' u 2:5 via o1,o2,o3,a
if ( B == 1 ) fit f(x) 'log' u 2:5 via o1,o2,a
!rm fit.log
set xr[0:4096]
if ( B >  1 ) fit f(x) 'log' u 2:5 via o1,o2,o3,a
if ( B == 1 ) fit f(x) 'log' u 2:5 via o1,o2,a
!rm fit.log
set xr[0:10000]
if ( B >  1 ) fit f(x) 'log' u 2:5 via o1,o2,o3,a
if ( B == 1 ) fit f(x) 'log' u 2:5 via o1,o2,a
EOS_0
}

create_AT_sampler2()
{
	init_IN_file2

	touch exe-sgemvt-sub2
	\rm exe-sgemvt-sub2

	cat << EOS_S > exe-sgemvt-sub2
#!/bin/bash

export BLOCK_SIZE=\$1
export UMAX=\$2
export VMAX=\$3
export UNROLL=\$4
export IN_DATA=\$5
export KERNEL_IDX=\$6

touch sgemvt-at-opt.c
\rm sgemvt-at-opt.c

cat << EOSS_SS >> sgemvt-at-opt.c
#define KERNEL_IDX \$KERNEL_IDX
#define BLOCK_SIZE \$BLOCK_SIZE
#define UMAX       \$UMAX
#define VMAX       \$VMAX
#define UNROLL     \$UNROLL

int
do_best_kernel_SGEMVT(int blk, const scalar_t alpha, const scalar_t beta,
        const scalar_t *a, const scalar_t *x, scalar_t *y,
        int lda, int n1, int n2, int offsetX, double *ff)
{
	GEMVT_KERNEL < GPU_ARCH, scalar_t, BLOCK_SIZE, VMAX, UMAX, UNROLL >
	        ( alpha, beta, a, x, y, lda, n1, n2, offsetX );
	return 0;
}
#undef BLOCK_SIZE
#undef UMAX
#undef VMAX
#undef UNROLL
EOSS_SS

cd ../src >& /dev/null
\rm sgemv_t.cu_o >& /dev/null
\rm sgemv_t-*.cu_o >& /dev/null
make -j4 >& /dev/null
cd ../bench >& /dev/null
\rm test-f.o >& /dev/null
make test-s >& /dev/null
cd ../tuning

unset ASPEN_VERIFY
$EXEC $BENCHCODE \$IN_DATA | awk '/^N=/{print;fflush();}' > logs/log-\$BLOCK_SIZE-\$UMAX-\$VMAX-\$UNROLL

if [ 1 -eq 1 ]; then

LOG=logs/log-MYSGEMV-T-\$BLOCK_SIZE"x"\$UMAX"x"\$VMAX"x"\$UNROLL-X-GTX560
CY=\`awk '/CYCLE/{ print \$2 }' \$LOG\`
echo 'CYCLE= '\$CY >> logs/log-\$BLOCK_SIZE-\$UMAX-\$VMAX-\$UNROLL

else
if [ x\${7} != 'x-' ]; then
	touch log-L
	\rm log-L
	MP=\$7
	i0=\`expr \$UMAX "*" \$VMAX\`
	i0=\`expr \$i0 "*" \$MP\`
	if [ \$i0 -le 2559 ]; then
		i1=\`expr 2559 "/" \$i0\`
		i2=\`expr \$i1 "+" 1\`
		i1=\`expr \$i2 "*" \$i0\`
	else
		i1=\$i0
	fi
	i1=0;
	
	for MM in \
		$MM_PATTERN
	do
		echo 'MM= '\$MM >> log-L
		awk 'BEGIN{\
			UMAX='\$UMAX';\\
			VMAX='\$VMAX';\\
			MP='\$MP'; \\
			MM='\$MM'; \\
			; \\
			if(MM<='$MAX_MULTI'){\\
				i2=UMAX*VMAX*MP*MM;\\
				i1=int((2560-1)/i2)+1;\\
				i0=i1*i2;\\
			} else {\\
				i2=UMAX*VMAX*MP*(MM-'$MAX_MULTI');\\
				i1=int((2560-1)/i2)+1;\\
				i0=(i1==1)?(i1*i2):((i1+1)*i2);\\
				i0=((i1+1)*i2);\\
			} \\
			; \\
			print 1;\\
			for(i=i0-32;i<=i0-32+8;i++)print i;\\
			print -1;\\
			}\\
		' > IN-b
		unset ASPEN_VERIFY
		$EXEC $BENCHCODE IN-b > log-a-L
		\rm IN-b
		awk 'BEGIN{\\
			UMAX='\$UMAX';\\
			VMAX='\$VMAX';\\
			MP='\$MP'; \\
			MM='\$MM'; \\
			; \\
			i2=UMAX*VMAX*MP*MM;\\
			if(MM<='$MAX_MULTI'){\\
				i2=UMAX*VMAX*MP*MM;\\
				i1=int((2560-1)/i2)+1;\\
				i0=i1*i2;\\
			} else {\\
				i2=UMAX*VMAX*MP*(MM-'$MAX_MULTI');\\
				i1=int((2560-1)/i2)+1;\\
				i0=(i1==1)?(i1*i2):((i1+1)*i2);\\
				i0=((i1+1)*i2);\\
			} \\
			; \\
			print 1;\\
			for(i=i0+32-8;i<=i0+32;i++)print i;\\
			print -1;\\
			}\\
		' > IN-b
		unset ASPEN_VERIFY
		$EXEC $BENCHCODE IN-b > log-a-R
		\rm IN-b
		cat log-a-L \\
		| awk 'BEGIN{s=0;i=0;}/N=/{if(s<\$5)s=\$5;i++;}END{print s;}' > log-L-L
		cat log-a-R \\
		| awk 'BEGIN{s=0;i=0;}/N=/{if(s<\$5)s=\$5;i++;}END{print s;}' > log-L-R
		cat log-L-L log-L-R | awk '{ a=\$0; getline; b=\$0; \\
			UMAX='\$UMAX'; \\
			VMAX='\$VMAX'; \\
			MP='\$MP'; \\
			MM='\$MM'; \\
			; \\
			if(MM<='$MAX_MULTI'){\\
				i2=UMAX*VMAX*MP*MM;\\
				i1=int((2560-1)/i2)+1;\\
				i0=i1*i2;\\
			} else {\\
				i2=UMAX*VMAX*MP*(MM-'$MAX_MULTI');\\
				i1=int((2560-1)/i2)+1;\\
				i0=(i1==1)?(i1*i2):((i1+1)*i2);\\
				i0=((i1+1)*i2);\\
			} \\
			; \\
			print \\
			i2" "i0 \\
			" "a" "b" "(a-b)/a; }'
		cat log-L-L log-L-R >> log-L
		\rm log-a-L log-a-R
		\rm log-L-L log-L-R
	done
	cat log-L \\
	| awk '/MM=/{ \\
		MM=\$2; getline; L[MM]=\$0; getline; R[MM]=\$0; \\
		G[MM]=(L[MM]-R[MM])/L[MM]; \\
		; \\
			UMAX='\$UMAX'; \\
			VMAX='\$VMAX'; \\
			MP='\$MP'; \\
			; \\
			i2=UMAX*VMAX*MP*MM; \\
			i1=int((2560-1)/i2)+1; \\
			i0=i1*i2; \\
		X[MM]=i0; \\
	}END{ \\
		MM=0; \\
		for(kk=1;kk<='$MAX_MULTI';kk++) if ( MM == 0 ) { \\
			for(k=kk;k<='$MAX_MULTI';k+=kk){ \\
				if ( G[k] <= 0.005 || G[k+'$MAX_MULTI'] <= 0.005 )  break; \\
			} \\
			if ( k > '$MAX_MULTI' ) { \\
				s = G[kk+'$MAX_MULTI']; \\
				for(k=2*kk;k<='$MAX_MULTI';k+=kk){ \\
					if ( s * 1.5 < G[k+'$MAX_MULTI'] ) s = G[k+'$MAX_MULTI']; \\
				} \\
				if ( s != G[kk+'$MAX_MULTI'] ) { \\
					MM = 0; \\
				} else {\\
					MM = kk; \\
				} \\
			} \\
		} \\
		NN=0; \\
		for(kk=MM+1;kk<='$MAX_MULTI';kk++) if ( NN == 0 ) { \\
			for(k=kk;k<='$MAX_MULTI';k+=kk){ \\
				if ( G[k] <= 0.005 || G[k+'$MAX_MULTI'] <= 0.005 )  break; \\
			} \\
			if ( k > '$MAX_MULTI' ) { \\
				s = G[kk+'$MAX_MULTI']; \\
				for(k=2*kk;k<='$MAX_MULTI';k+=kk){ \\
					if ( s * 1.5 < G[k+'$MAX_MULTI'] ) s = G[k+'$MAX_MULTI']; \\
				} \\
				if ( s != G[kk+'$MAX_MULTI'] ) { \\
					NN = 0; \\
				} else { \\
					NN = kk; \\
				} \\
			} \\
		} \\
		if ( NN > 0 ) { \\
			s = ( NN < 2*MM ) ? 1.05 : 1.5; \\
			if ( G[NN] > G[MM] * s ) MM = NN; \\
		} \\
	if (0) {\\
		tj=0; \\
		for(k=1;k<='$MAX_MULTI';k++){ \\
			if ( G['$MAX_MULTI'+k] > 0.0 ) { tt[tj++]=k; } \\
		} \\
		j = -1; s = X[j] = G[j] = 0; \\
		B = '\$UMAX'*'\$VMAX'*'\$MP'; \\
		for(i=0;i<tj;i++){ \\
			if ( ( X[j] == X[tt[i]] && tt[i] % j != 0 ) ) { \\
				print B*j" vs "B*tt[i];\\
				print G[j+'$MAX_MULTI']" vs "G[tt[i]+'$MAX_MULTI'];\\
				if ( G[j+'$MAX_MULTI'] < 0.5*G[tt[i]+'$MAX_MULTI'] ) { \\
					s = G[j=tt[i]]; \\
				} \\
			} else if ( s <= G[tt[i]+'$MAX_MULTI'] ) { \\
				if ( G[tt[i]] > 0.0 && \\
					0.5*G[tt[i]] < G[tt[i]+'$MAX_MULTI'] ) { \\
					s = G[(j=tt[i])+'$MAX_MULTI']; \\
				} \\
			} \\
		} \\
		if ( s < 0.02 ) { \\
			for(k=1;k<='$MAX_MULTI';k++){\\
				if ( G[k]<.005 || G[k+'$MAX_MULTI']<.005 ) break; \\
			}\\
			if ( k > '$MAX_MULTI' ) j = 1; \\
		}\\
		if ( j < 0 ) j = 0; \\
		MM=j; \\
	}\\
		print "CYCLE= "'\$UMAX'*'\$VMAX'*'\$MP'*MM; \\
	}' | tee -a logs/log-\$BLOCK_SIZE-\$UMAX-\$VMAX-\$UNROLL
	\rm log-L
fi
fi
EOS_S
}

do_Fitting()
{
	create_GP_plot

	cat at.data | \
	awk ' \
		/DIAG/{a=1;next;} \
		{if(a)print;} \
	' | \
	awk ' BEGIN{IDX=0} \
		/^[0-9]/{ IDX++; \
			split($2,aa,":"); \
			#system("/bin/sh exe-sgemvt-sub2 "aa[1]" "aa[2]" "aa[3]" "aa[4]" '$IN_DATA' 64 '$MP' - >& /dev/null"); \
			system("cp logs/log-"aa[1]"-"aa[2]"-"aa[3]"-"aa[4]" log"); \
			printf("MP="'$MP'"; UMAX="aa[2]"; VMAX="aa[3]";\n") > "param.plt"; \
			close("param.plt"); \
			system("echo CYCLE=1 >> param.plt"); \
			system("grep CYCLE log >> param.plt"); \
			system("touch fit.log; rm -f fit.log; gnuplot bb.plot >& /dev/null"); \
			f = "fit.log"; \
			while( getline<f ) { \
				if($0~/Final/){ \
					o1=o2=o3=a=0;
					while(1) { \
						getline<f; \
						if ( $0~/o1[ ]*=/ ) o1=$3; \
						if ( $0~/o2[ ]*=/ ) o2=$3; \
						if ( $0~/o3[ ]*=/ ) o3=$3; \
						if ( $0~/a[ ]*=/ ) a=$3; \
						if ( a != 0 ) break; \
					} \
				} \
			} close(f); \
			print "### "aa[1]" "aa[2]" "aa[3]" "aa[4]" "o1" "o2" "o3" "a; \
			system("grep CYCLE log");
		} \
	' | \
	awk ' \
		/###/{print} \
		/CYCLE/{print} \
	' | tee bb.data

	FIT_INIT=`awk ' \
		BEGIN{i=0}\
		/^###/{\
			i++; o1+=$6; o2+=$7; o3+=$8; a+=$9; \
		}\
		END{\
			o1/=i; o2/=i; o3/=i; a/=i; \
			o1*=1.0001; o2*=1.0001; o3*=1.0001; a*=1.0001; \
			print "o1="o1";o2="o2";o3="o3";a="a";"}' \
		bb.data`

	\rm log
}

generate_AT_API()
{

	echo '#if 0'
	echo_Message API
	echo '#endif'

        echo "#if 0"
        echo 'char* _AT_tuned_host="'$HOST'";'
	echo 'char* _AT_tuned_device="'$DEVICE'";'
        echo "#endif"

	cat << EOS_1
#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#if !defined(API)
#define API(x)			x
#define get_numbers_MP(...)	($MP)
#endif
#if !defined(scalar_t)
#define scalar_t	float
#endif

#define _NAME_     _NAME_4PARAMS_

#if __SIMULATE__
static int	kernel_id;
#endif

long
do_best_kernel_SGEMVT(int blk, const scalar_t alpha, const scalar_t beta,
	const scalar_t *a, const scalar_t *x, scalar_t *y,
	int lda, int n1, int n2, int offsetX, double *ff)
{
	struct _t { long id; double a; double o1; double o2; double o3; int cycle; }
	model[] = {
EOS_1

	awk ' \
	BEGIN{i=0} \
	/^###/{ \
		BLOCK_SIZE[i]=$2; UMAX[i]=$3; VMAX[i]=$4; UNROLL[i]=$5; \
		o[i]=$6; o2[i]=$7; o3[i]=$8; a[i]=$9; i++; \
	} \
	/CYCLE/{ CYCLE[i-1]=$2; }\
	END{ N=i; \
		for(j=0;j<N;j++) {\
			printf("\t\t // KERNEL_IDX=%d  BLOCK_SIZE=%d  UMAX=%d  VMAX=%d  UNROLL=%d  CYCLE=%d\n", \
				j, BLOCK_SIZE[j], UMAX[j], VMAX[j], UNROLL[j], CYCLE[j]); \
			printf("\t\t{ 0x%010x, %10g, %10g, %10g, %10g, %d },\n", \
				lshift(BLOCK_SIZE[j], 0)+lshift(UMAX[j],16)+lshift(VMAX[j],24)+lshift(UNROLL[j],32), \
				a[j], o[j], o2[j], o3[j], CYCLE[j]); \
		} \
	} \
	' bb.data

	cat << EOS_2
		{ 0, 0.0, 0.0, 0.0, 0.0, 0 }
	};

	int  MP = API(get_numbers_MP)();

	int	i, ii = 0; long t=0;
if ( blk < 0 ) {
	int	n = n2;
	double X=(double)n, fx=0.0f;
	for(i=0;model[i].id>0;i++){
		double a=(double)model[i].a;
		double o1=(double)model[i].o1;
		double o2=(double)model[i].o2;
		double o3=(double)model[i].o3;
		// int BLOCK_SIZE=((model[i].id>>0)&0xffff);
		// int UMAX=((model[i].id>>16)&0xff);
		// int VMAX=((model[i].id>>24)&0xff);
		// int UNROLL=((model[i].id>>32)&0xff);
		int B; B=model[i].cycle; if ( B <= 0 ) B=1;
		if ($CG<200) {
			if ( B > 1 && B <= 1500 && B > 11*MP ) {
                        	if ( n % B < MP*8 ) continue;
	                        if ( n % B > B-MP*3 ) continue;
			}
		}
		double f=a*(X*X)/((o1*o1)+(o2+o3*floor(X/B)*B+(1-o3)*X)*X);

//if ( B > 3600 || X < B * (MP-4)/MP || X > B )
		if(f>fx){ t=model[i].id; ii=i; fx=f; }
	} *ff=fx;
} else {
	int	n = n2;
	double X=(double)n;
	i = blk;
		double a=(double)model[i].a;
		double o1=(double)model[i].o1;
		double o2=(double)model[i].o2;
		double o3=(double)model[i].o3;
		int B; B=model[i].cycle; if ( B <= 0 ) B=1;
		double f=a*(X*X)/((o1*o1)+(o2+o3*floor(X/B)*B+(1-o3)*X)*X);
	ii = blk;
	*ff=f;
}

#if __SIMULATE__
	kernel_id = ii;
#else

        #define SGEMVT_KERNEL_(BLOCK_SIZE,VMAX,UMAX,UNROLL) \\
                GEMVT_KERNEL < GPU_ARCH, scalar_t, BLOCK_SIZE, VMAX, UMAX, UNROLL > \\
                ( alpha, beta, a, x, y, lda, n1, n2, offsetX )

	switch (ii) {
EOS_2

	awk '\
	BEGIN{i=0} \
	/^###/{ \
		BLOCK_SIZE[i]=$2; UMAX[i]=$3; VMAX[i]=$4; UNROLL[i]=$5; \
		o[i]=$6; o2[i]=$7; o3[i]=$8; a[i]=$9; i++; \
	} \
	END{ N=i; \
		for(j=0;j<N;j++) { \
			print "\tcase "j":"; \
			print "\t\tSGEMVT_KERNEL_ ( "BLOCK_SIZE[j]" , "VMAX[j]" , "UMAX[j]", "UNROLL[j]" );";\
			print "\t\tbreak;";\
		} \
	} \
	' bb.data

	cat << EOS_3
	default:
		fprintf(stderr,"Illegal BLK parameter\n");
		break;
	}
#endif
	return t;
}

#if __SIMULATE__
main(int argc, char **argv){
  int blk=-1;
  if ( argc > 1 ) blk = atoi(argv[1]);
  while(1){
    int n; scanf("%d", &n); if(n<1)break;
    const double *x = NULL;
    double *y = NULL;
    double s; long k=do_best_kernel_SGEMVT(blk, 0.0, 0.0, x, x, y, 0, n, n, 0, &s);
    printf("%d BLK_SIZE=%d UMAX=%d VMAX=%d UNROLL=%d %g %d\n",
    n, ((k>>0)&0x0ffff), ((k>>16)&0xff), ((k>>24)&0xff), ((k>>32)&0xff), s,
	kernel_id);
  }
}
#endif
EOS_3
}

epilog()
{
	if [ -f sgemvt-at-opt.c ]; then
		cp sgemvt-at-opt.c logs/
	fi
	if [ -f sgemvt-at-opt1.c ]; then
		cp sgemvt-at-opt1.c logs/
	fi
	if [ -f sgemvt-at-opt2.h ]; then
		cp sgemvt-at-opt2.h logs/
	fi
	if [ -f sgemvt-at-opt3.h ]; then
		cp sgemvt-at-opt3.h logs/
	fi

	if [ -d logs ]; then
		LOG=log-sgemvt-AT1X.`date +%Y.%m.%d.%T`
		mv logs $LOG
	fi
	if [ -f $IN_DATA ]; then
		\rm $IN_DATA
	fi
	if [ -f at-log ]; then
		\rm at-log
	fi
	if [ -f at.data ]; then
		\rm at.data
	fi
	if [ -f bb.data ]; then
		\rm bb.data
	fi
	if [ -f bb.plot ]; then
		\rm bb.plot
	fi
	if [ -f param.plt ]; then
		\rm param.plt
	fi
	if [ -f log ]; then
		\rm log
	fi
	if [ -f fit.log ]; then
		\rm fit.log
	fi
	if [ -f exe-sgemvt-sub2 ]; then
		\rm exe-sgemvt-sub2
	fi
	if [ -f SdkMasterLog.csv ]; then
		\rm SdkMasterLog.csv
	fi
	if [ -f deviceQuery.txt ]; then
		\rm deviceQuery.txt
	fi

	if [ ! -d data ]; then
		mkdir data
	fi
	if [ ! -d data/sgemvt ]; then
		mkdir data/sgemvt
	fi
	if [ ! -d data/sgemvt/$DEVICE ]; then
		mkdir data/sgemvt/$DEVICE
	fi
	cp sgemvt-at-opt.c data/sgemvt/$DEVICE/$HOST-`date +%Y%m%d.%H%M%S`
}



prolog

echo_Message begin

if [ x${CUDA_CHECK} = xFAILED ]; then
	echo '*** WARNING ***'
	echo 'CUDA_SDK_PATH does not match the one specified on sgemvn.sh'
	echo 'PLEASE correct CUDA_SDK_PATH.'
	exit 1
fi

initial_Phase

do_AT_sampler | tee at-log

do_Ranking

do_AT_sampler2

for itr in 1 2 3 4 5
do
        echo 'Fitting '$itr'-time'
        do_Fitting
done

generate_AT_API > sgemvt-at-opt1.c

/bin/sh ./cmvts.sh

cat sgemvt-at-opt.c

update_MYBLAS

epilog

echo_Message end

touch .done-sgemvt

exit 0

